from hypers._tools._smoothen import _data_smoothen
from hypers._tools._types import PreprocessType, ClusterType, DecomposeType, MixtureType
from hypers._tools._update import _data_mean, _data_checks, _data_access

__all__ = ['_data_smoothen', '_data_mean', '_data_checks', '_data_access', 'PreprocessType',
           'ClusterType', 'DecomposeType', 'MixtureType']



import hypers as hp
from scipy.signal import savgol_filter
from scipy.ndimage.filters import gaussian_filter


def _data_smoothen(X: 'hp.Dataset', **kwargs):

    if X.smoothing == 'savitzky_golay':
        if X.ndim == 3:
            for xpix in range(X.shape[0]):
                for ypix in range(X.shape[1]):
                    X.data[xpix, ypix, :] = savgol_filter(X.data[xpix, ypix, :], **kwargs)

        elif X.ndim == 4:
            for xpix in range(X.shape[0]):
                for ypix in range(X.shape[1]):
                    for zpix in range(X.shape[2]):
                        X.data[xpix, ypix, zpix, :] = savgol_filter(X.data[xpix, ypix, zpix, :], **kwargs)

    elif X.smoothing == 'gaussian_filter':
        if X.ndim == 3:
            for xpix in range(X.shape[0]):
                for ypix in range(X.shape[1]):
                    X.data[xpix, ypix, :] = gaussian_filter(X.data[xpix, ypix, :], **kwargs)

        elif X.ndim == 4:
            for xpix in range(X.shape[0]):
                for ypix in range(X.shape[1]):
                    for zpix in range(X.shape[2]):
                        X.data[xpix, ypix, zpix, :] = gaussian_filter(X.data[xpix, ypix, zpix, :], **kwargs)



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mainwindow.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1243, 458)
        MainWindow.setStyleSheet("%background: #161616")
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.imagewin = ImageView(self.centralwidget)
        self.imagewin.setObjectName("imagewin")
        self.horizontalLayout_2.addWidget(self.imagewin)
        self.specwin = PlotWidget(self.centralwidget)
        self.specwin.setObjectName("specwin")
        self.horizontalLayout_2.addWidget(self.specwin)
        self.horizontalLayout_2.setStretch(0, 40)
        self.horizontalLayout_2.setStretch(1, 60)
        self.gridLayout.addLayout(self.horizontalLayout_2, 1, 0, 1, 1)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.slider = QtWidgets.QSlider(self.centralwidget)
        self.slider.setMaximum(10)
        self.slider.setPageStep(1)
        self.slider.setOrientation(QtCore.Qt.Horizontal)
        self.slider.setObjectName("slider")
        self.horizontalLayout_3.addWidget(self.slider)
        self.updateSpectrum = QtWidgets.QPushButton(self.centralwidget)
        self.updateSpectrum.setObjectName("updateSpectrum")
        self.horizontalLayout_3.addWidget(self.updateSpectrum)
        self.updateImage = QtWidgets.QPushButton(self.centralwidget)
        self.updateImage.setObjectName("updateImage")
        self.horizontalLayout_3.addWidget(self.updateImage)
        self.Reset = QtWidgets.QPushButton(self.centralwidget)
        self.Reset.setObjectName("Reset")
        self.horizontalLayout_3.addWidget(self.Reset)
        self.gridLayout.addLayout(self.horizontalLayout_3, 2, 0, 1, 1)
        self.gridLayout.setRowStretch(1, 95)
        self.gridLayout.setRowStretch(2, 5)
        MainWindow.setCentralWidget(self.centralwidget)

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "View Hyperspectral Image"))
        self.updateSpectrum.setText(_translate("MainWindow", " Update Spectrum "))
        self.updateImage.setText(_translate("MainWindow", " Update Image "))
        self.Reset.setText(_translate("MainWindow", "RESET"))

from pyqtgraph import ImageView, PlotWidget



import numpy as np
import hypers as hp
from typing import Tuple
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from sklearn.cluster import AffinityPropagation, MeanShift, DBSCAN
from hypers._tools._types import ClusterType, CLUSTER_TYPES


def _n_clusters(X: 'hp.Dataset',
                mdl: ClusterType) -> int:

    if not type(mdl) in (AffinityPropagation, MeanShift, DBSCAN):
        n_clusters = X.mdl_cluster.get_params()['n_clusters']

    elif type(mdl) == AffinityPropagation:
        n_clusters = X.mdl_cluster.cluster_centers_indices_.shape[0]

    elif type(mdl) == MeanShift:
        n_clusters = X.mdl_cluster.cluster_centers_.shape[0]

    elif type(mdl) == DBSCAN:
        n_clusters = X.mdl_cluster.core_sample_indices_.shape[0]

    else:
        raise TypeError

    return n_clusters


def _data_cluster(X: 'hp.Dataset',
                  mdl: ClusterType,
                  decomposed: bool,
                  pca_comps: int,
                  plot: bool,
                  return_arrs: bool) -> Tuple[np.ndarray, np.ndarray]:

    if type(mdl) not in CLUSTER_TYPES:
        raise TypeError('Must pass a sklearn cluster class. Refer to documentation.')

    X.mdl_cluster = mdl

    if decomposed:
        print('Clustering with the first ' + str(pca_comps) + ' PCA components.')
        mdl_pca = PCA(n_components=pca_comps)
        comps = mdl_pca.fit_transform(X.flatten())
        X.mdl_cluster.fit(comps)
        n_clusters = _n_clusters(X, mdl)
        if type(X.mdl_cluster) in (AffinityPropagation, MeanShift):
            n_clusters = X.mdl_cluster.cluster_centers_.shape[0]
        labels = X.mdl_cluster.labels_.reshape(X.data.shape[:-1])

    else:
        X.mdl_cluster.fit(X.flatten())
        n_clusters = _n_clusters(X, mdl)
        labels = X.mdl_cluster.labels_.reshape(X.data.shape[:-1])

    try:
        if decomposed:
            specs = mdl_pca.inverse_transform(X.mdl_cluster.cluster_centers_)
        else:
            specs = X.mdl_cluster.cluster_centers_

    except AttributeError:
        specs = np.zeros((n_clusters, X.data.shape[-1]))
        lbls = labels + 1
        for cluster_number in range(n_clusters):
            msk = np.zeros(X.data.shape)
            for spectral_point in range(X.data.shape[-1]):
                msk[..., spectral_point] = np.multiply(
                    X.data[..., spectral_point],
                    np.where(lbls == cluster_number+1, lbls, 0)/(cluster_number+1)
                )
                    
            if X.ndim == 3:
                specs[cluster_number, :] = np.squeeze(np.mean(np.mean(msk, 1), 0))
            elif X.ndim == 4:
                specs[cluster_number, :] = np.squeeze(np.mean(np.mean(np.mean(msk, 2), 1), 0))

    specs = specs.T

    if plot:
        plt.subplot(121)
        if X.ndim == 3:
            plt.imshow(labels)
        elif X.ndim == 4:
            plt.imshow(np.squeeze(np.mean(labels, -1)))
        plt.axis('off')
        plt.subplot(122)
        for cluster_number in range(n_clusters):
            plt.plot(specs[..., cluster_number])
        plt.show()

    if return_arrs:
        return labels, specs



import numpy as np
import hypers as hp
from typing import Tuple
import matplotlib.pyplot as plt
from sklearn.decomposition import PCA
from hypers._tools._types import DecomposeType, DECOMPOSE_TYPES


def _data_decomposition(X: 'hp.Dataset',
                        mdl: DecomposeType,
                        plot: bool,
                        return_arrs: bool) -> Tuple[np.ndarray, np.ndarray]:

    if type(mdl) not in DECOMPOSE_TYPES:
        raise TypeError('Must pass a sklearn decomposition class. Refer to documentation.')

    X.mdl_decompose = mdl
    n_components = X.mdl_decompose.get_params()['n_components']
    images = X.mdl_decompose.fit_transform(X.flatten()).reshape(X.data.shape[:-1] + (n_components,))
    specs = X.mdl_decompose.components_.transpose()

    if plot:
        for component in range(n_components):
            plt.subplot(n_components, 2, 2*component + 1)
            if X.ndim == 3:
                plt.imshow(np.squeeze(images[..., component]))
            elif X.ndim == 4:
                plt.imshow(np.mean(np.squeeze(images[..., component]), -1))
            plt.axis('off')
            plt.subplot(n_components, 2, 2 * component + 2)
            plt.plot(specs[..., component])
        plt.show()

    if return_arrs:
        return images, specs


def _data_scree(X: 'hp.Dataset',
                plot: bool,
                return_arrs: bool) -> np.ndarray:

    mdl = PCA()
    mdl.fit_transform(X.flatten())
    scree = mdl.explained_variance_ratio_

    X._scree = scree

    if plot:
        plt.plot(scree)
        plt.xlabel('Principal components')
        plt.ylabel('Variance ratio')
        plt.title('PCA scree plot')
        plt.show()

    if return_arrs:
        return scree



import numpy as np
import hypers as hp
from typing import Tuple, List
import matplotlib.pyplot as plt


def _vca(X: 'hp.Dataset',
         n_components: int = 4,
         input_snr: float = 0,
         plot: bool = False,
         return_arrs: bool = True) -> Tuple[np.ndarray, List[int]]:

    Ae, indice, Yp = _calcluate_vca(X.flatten().T, n_components, snr_input=input_snr)
    index = [0]*n_components
    for component in range(n_components):
        index[component] = np.unravel_index(indice[component], X.shape[:-1])

    if plot:
        for component in range(n_components):
            plt.plot(Ae[:, component])
        plt.show()

    if return_arrs:
        return Ae, index


def _estimate_snr(Y: np.ndarray,
                  r_m: np.ndarray,
                  x: np.ndarray) -> np.ndarray:

    [L, N] = Y.shape  # L number of bands (channels), N number of pixels
    [p, N] = x.shape  # p number of endmembers (reduced dimension)

    P_y = np.sum(Y ** 2) / float(N)
    P_x = np.sum(x ** 2) / float(N) + np.sum(r_m ** 2)
    snr_est = 10 * np.log10((P_x - p / L * P_y) / (P_y - P_x))

    return snr_est


def _calcluate_vca(Y, R, verbose=True, snr_input=0):
    # Vertex Component Analysis
    #
    # Ae, indice, Yp = vca(Y,R,verbose = True,snr_input = 0)
    #
    # ------- Input variables -------------
    #  Y - matrix with dimensions L(channels) x N(pixels)
    #      each pixel is a linear mixture of R endmembers
    #      signatures Y = M x s, where s = gamma x alfa
    #      gamma is a illumination perturbation factor and
    #      alfa are the abundance fractions of each endmember.
    #  R - positive integer number of endmembers in the scene
    #
    # ------- Output variables -----------
    # Ae     - estimated mixing matrix (endmembers signatures)
    # indice - pixels that were chosen to be the most pure
    # Yp     - Data matrix Y projected.
    #
    # ------- Optional parameters---------
    # snr_input - (float) signal to noise ratio (dB)
    # v         - [True | False]
    # ------------------------------------
    #
    # Author: Adrien Lagrange (adrien.lagrange@enseeiht.fr)
    # This code is a translation of a matlab code provided by
    # Jose Nascimento (zen@isel.pt) and Jose Bioucas Dias (bioucas@lx.it.pt)
    # available at http://www.lx.it.pt/~bioucas/code.htm under a non-specified Copyright (c)
    # Translation of last version at 22-February-2018 (Matlab version 2.1 (7-May-2004))
    #
    # more details on:
    # Jose M. P. Nascimento and Jose M. B. Dias
    # "Vertex Component Analysis: A Fast Algorithm to Unmix Hyperspectral Data"
    # submited to IEEE Trans. Geosci. Remote Sensing, vol. .., no. .., pp. .-., 2004
    #
    #

    #############################################
    # Initializations
    #############################################
    if len(Y.shape) != 2:
        raise ValueError('Input data must be of size L (number of bands i.e. channels) by N (number of pixels)')

    [L, N] = Y.shape  # L number of bands (channels), N number of pixels

    R = int(R)
    if R < 0 or R > L:
        raise ValueError('ENDMEMBER parameter must be integer between 1 and L')

    #############################################
    # SNR Estimates
    #############################################

    if snr_input == 0:
        y_m = np.mean(Y, axis=1, keepdims=True)
        Y_o = Y - y_m  # data with zero-mean
        Ud = np.linalg.svd(np.dot(Y_o, Y_o.T) / float(N))[0][:, :R]  # computes the R-projection matrix
        x_p = np.dot(Ud.T, Y_o)  # project the zero-mean data onto p-subspace

        SNR = _estimate_snr(Y, y_m, x_p);

        if verbose:
            print("SNR estimated = {}[dB]".format(SNR))
    else:
        SNR = snr_input
        if verbose:
            print("input SNR = {}[dB]\n".format(SNR))

    SNR_th = 15 + 10 * np.log10(R)

    #############################################
    # Choosing Projective Projection or
    #          projection to p-1 subspace
    #############################################

    if SNR < SNR_th:
        if verbose:
            print("... Select proj. to R-1")

            d = R - 1
            if snr_input == 0:  # it means that the projection is already computed
                Ud = Ud[:, :d]
            else:
                y_m = np.mean(Y, axis=1, keepdims=True)
                Y_o = Y - y_m  # data with zero-mean

                Ud = np.linalg.svd(np.dot(Y_o, Y_o.T) / float(N))[0][:, :d]  # computes the p-projection matrix
                x_p = np.dot(Ud.T, Y_o)  # project thezeros mean data onto p-subspace

            Yp = np.dot(Ud, x_p[:d, :]) + y_m  # again in dimension L

            x = x_p[:d, :]  # x_p =  Ud.T * Y_o is on a R-dim subspace
            c = np.amax(np.sum(x ** 2, axis=0)) ** 0.5
            y = np.vstack((x, c * np.ones((1, N))))
    else:
        if verbose:
            print("... Select the projective proj.")

        d = R
        Ud = np.linalg.svd(np.dot(Y, Y.T) / float(N))[0][:, :d]  # computes the p-projection matrix

        x_p = np.dot(Ud.T, Y)
        Yp = np.dot(Ud, x_p[:d, :])  # again in dimension L (note that x_p has no null mean)

        x = np.dot(Ud.T, Y)
        u = np.mean(x, axis=1, keepdims=True)  # equivalent to  u = Ud.T * r_m
        y = x / np.dot(u.T, x)

    #############################################
    # VCA algorithm
    #############################################

    indice = np.zeros((R), dtype=int)
    A = np.zeros((R, R))
    A[-1, 0] = 1

    for i in range(R):
        w = np.random.rand(R, 1);
        f = w - np.dot(A, np.dot(np.linalg.pinv(A), w))
        f = f / np.linalg.norm(f)

        v = np.dot(f.T, y)

        indice[i] = np.argmax(np.absolute(v))
        A[:, i] = y[:, indice[i]]  # same as x(:,indice(i))

    Ae = Yp[:, indice]

    return Ae, indice, Yp



from typing import Union
from sklearn.mixture import GaussianMixture, BayesianGaussianMixture
from sklearn.preprocessing import (
    MaxAbsScaler, MinMaxScaler, PowerTransformer, QuantileTransformer, RobustScaler,
    StandardScaler, Normalizer
)
from sklearn.cluster import (
    KMeans, AffinityPropagation, MeanShift, SpectralClustering, AgglomerativeClustering, DBSCAN
)
from sklearn.decomposition import (
    PCA, FastICA, IncrementalPCA, TruncatedSVD, DictionaryLearning, MiniBatchDictionaryLearning,
    FactorAnalysis, NMF, LatentDirichletAllocation
)

PREPROCESSING_TYPES = (
    MaxAbsScaler,
    MinMaxScaler,
    PowerTransformer,
    QuantileTransformer,
    RobustScaler,
    StandardScaler,
    Normalizer
)

CLUSTER_TYPES = (
    KMeans,
    AffinityPropagation,
    MeanShift,
    SpectralClustering,
    AgglomerativeClustering,
    DBSCAN
)

DECOMPOSE_TYPES = (
    PCA,
    FastICA,
    #KernelPCA,
    IncrementalPCA,
    TruncatedSVD,
    DictionaryLearning,
    MiniBatchDictionaryLearning,
    FactorAnalysis,
    NMF,
    LatentDirichletAllocation
)

MIXTURE_TYPES = (
    GaussianMixture,
    BayesianGaussianMixture
)


PreprocessType = Union[
    MaxAbsScaler, MinMaxScaler, PowerTransformer, QuantileTransformer, RobustScaler, StandardScaler, Normalizer
]

ClusterType = Union[
    KMeans, AffinityPropagation, MeanShift, SpectralClustering, AgglomerativeClustering, DBSCAN
]

DecomposeType = Union[
    PCA, FastICA, IncrementalPCA, TruncatedSVD, DictionaryLearning, MiniBatchDictionaryLearning, NMF,
    FactorAnalysis, LatentDirichletAllocation
]

MixtureType = Union[
    GaussianMixture, BayesianGaussianMixture
]



from hypers._learning._cluster import _data_cluster
from hypers._learning._vca import _vca
from hypers._learning._abundance import _ucls
from hypers._learning._mixture import _data_mixture
from hypers._learning._decomposition import _data_decomposition, _data_scree

__all__ = ['_data_cluster', '_vca', '_ucls', '_data_decomposition', '_data_scree', '_data_mixture']



import numpy as np
import hypers as hp
import matplotlib.pyplot as plt


def _ucls(X: 'hp.Dataset',
          spectra: np.ndarray,
          plot: bool = False,
          return_arrs: bool = True) -> np.ndarray:

    x_inverse = np.linalg.pinv(spectra)
    im = np.dot(x_inverse, X.flatten().T).T.reshape(X.shape[:-1] + (spectra.shape[-1],))

    if plot:
        n_rows = np.ceil(spectra.shape[-1] / 3)
        if X.ndim == 3:
            for endm in range(spectra.shape[-1]):
                plt.subplot(n_rows, 3, endm+1)
                plt.imshow(np.squeeze(im[..., endm]))
                plt.title('Map ' + str(endm+1))
                plt.axis('off')
        elif X.ndim == 4:
            for endm in range(spectra.shape[-1]):
                plt.subplot(n_rows, 3, endm+1)
                plt.imshow(np.mean(np.squeeze(im[..., endm]), -1))
                plt.title('Map' + str(endm + 1))
                plt.axis('off')
        plt.show()

    if return_arrs:
        return im



import numpy as np
import hypers as hp
from typing import Tuple
import matplotlib.pyplot as plt
from hypers._tools._types import MixtureType, MIXTURE_TYPES


def _data_mixture(X: 'hp.Dataset',
                  mdl: MixtureType,
                  plot: bool,
                  return_arrs: bool) -> Tuple[np.ndarray, np.ndarray]:

    if type(mdl) not in MIXTURE_TYPES:
        raise TypeError('Must pass a sklearn mixture class. Refer to documentation.')

    X.mdl_mixture = mdl
    n_components = X.mdl_mixture.get_params()['n_components']
    lbls = X.mdl_mixture.fit_predict(X.flatten()).reshape(X.shape[:-1])
    spectra = X.mdl_mixture.means_.T

    if plot:
        plt.subplot(121)
        if X.ndim == 3:
            plt.imshow(lbls)
        elif X.ndim == 4:
            plt.imshow(np.mean(np.squeeze(lbls), -1))

        plt.subplot(122)
        for component in range(n_components):
            plt.plot(spectra[:, component])
        plt.show()

    if return_arrs:
        return lbls, spectra



import sys
import numpy as np
import pyqtgraph as pg
from PyQt5.Qt import QPalette, QColor
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QMainWindow, QApplication, QStyleFactory

import hypers as hp
from hypers._view._form import mainwindow


class HSIDialog(QMainWindow, mainwindow.Ui_MainWindow):
    """ Hyperspectral data viewer

    Displays the hyperspectral data. Features include:
    - Multiple layers (z-axis) by scrolling through the layers.
    - Viewing the image pertaining to chosen spectral bands.
    - Viewing the spectra averaged over a chosen region in the image.

    This class should not be imported directly. Instead either:
    - Import hsiPlot from skhyper.view and use on a Process object
    - Call the `view()` method of a Process object

    """
    def __init__(self, X: 'hp.Dataset', parent=None):
        super(HSIDialog, self).__init__(parent)
        self.setupUi(self)
        self.setWindowTitle('View Hyperspectral Data')

        if not isinstance(X, hp.Dataset):
            raise TypeError('Data needs to be passed to skhyper.process.Process first')

        self._X = X

        self.shape = None
        self.dimensions = None

        self.slider.valueChanged.connect(self.update_layer)
        self.updateImage.clicked.connect(self.update_image)
        self.updateSpectrum.clicked.connect(self.update_spectrum)
        self.Reset.clicked.connect(self.reset)

        # --- Setting image/plot settings -----------------------
        self.spec_lo = 0
        self.spec_hi = 0

        self.pgLRI = pg.LinearRegionItem()
        self.specwin.addItem(self.pgLRI)

        self.pgLRI.sigRegionChanged.connect(self.spec_region_updated)

        self.plotline = self.specwin.plot()
        # -------------------------------------------------------

        self.load_data()

    def load_data(self):
        if self._X is None:
            self.slider.setEnabled(False)

        else:
            self.shape = self._X.shape
            self.dimensions = self._X.ndim
            # self.slider.setMaximum(self.shape[2]-1)

            if self.dimensions == 3:
                self.slider.setEnabled(False)
                self.data_image(self._X.data)
                self.data_spectrum(self._X.data)

            elif self.dimensions == 4:
                self.slider.setValue(0)
                self.slider.setMaximum(self.shape[2]-1)
                self.data_image(self._X[:, :, 0, :])
                self.data_spectrum(self._X[:, :, 0, :])

    def spec_region_updated(self, regionItem):
        self.spec_lo, self.spec_hi = regionItem.getRegion()

    def update_layer(self):
        self.data_image(self._X[:, :, int(self.slider.value()), :])
        self.data_spectrum(self._X[:, :, int(self.slider.value()), :])

    def update_image(self):
        if self.dimensions == 3:
            self.data_image(self._X[:, :, int(self.spec_lo):int(self.spec_hi)])

        elif self.dimensions == 4:
            self.data_image(self._X[:, :, int(self.slider.value()) - 1, int(self.spec_lo):int(self.spec_hi)])

    def update_spectrum(self):
        # Obtaining coordinates of ROI graphic in the image plot
        image_coord_handles = self.imagewin.roi.getState()
        posimage = image_coord_handles['pos']
        sizeimage = image_coord_handles['size']

        posx = int(posimage[0])
        sizex = int(sizeimage[0])
        posy = int(posimage[1])
        sizey = int(sizeimage[1])
        xmin = posx
        xmax = posx + sizex
        ymin = posy
        ymax = posy + sizey

        if self.dimensions == 3:
            self.data_spectrum(self._X[xmin:xmax, ymin:ymax, :])

        elif self.dimensions == 4:
            self.data_spectrum(self._X[xmin:xmax, ymin:ymax, int(self.slider.value()) - 1, :])

    def reset(self):
        self.load_data()

    def data_image(self, data):
        self.imagewin.setImage(np.squeeze(np.mean(data, 2)))

    def data_spectrum(self, data):
        self.plotline.setData(np.squeeze(np.mean(np.mean(data, 1), 0)))


def hsiPlot(X):
    """Hyperspectral data viewer

    Displays the hyperspectral data. This is one of two methods to do this.

    Parameters
    ----------
    X : object, Process instance
        The object X containing the hyperspectral array.

    """
    app = QApplication(sys.argv)

    # Setting the dark-themed Fusion style for the GUI
    app.setStyle(QStyleFactory.create('Fusion'))
    dark_palette = QPalette()
    dark_palette.setColor(QPalette.Window, QColor(23, 23, 23))
    dark_palette.setColor(QPalette.WindowText, QColor(200, 200, 200))
    dark_palette.setColor(QPalette.Base, QColor(18, 18, 18))
    dark_palette.setColor(QPalette.AlternateBase, QColor(53, 53, 53))
    dark_palette.setColor(QPalette.ToolTipBase, Qt.white)
    dark_palette.setColor(QPalette.ToolTipText, Qt.white)
    dark_palette.setColor(QPalette.Text, QColor(200, 200, 200))
    dark_palette.setColor(QPalette.Button, QColor(33, 33, 33))
    dark_palette.setColor(QPalette.ButtonText, QColor(200, 200, 200))
    dark_palette.setColor(QPalette.BrightText, Qt.red)
    dark_palette.setColor(QPalette.Link, QColor(42, 130, 218))
    dark_palette.setColor(QPalette.Highlight, QColor(42, 130, 218))
    dark_palette.setColor(QPalette.HighlightedText, Qt.white)
    dark_palette.setColor(QPalette.Active, QPalette.Button, QColor(33, 33, 33))
    dark_palette.setColor(QPalette.Disabled, QPalette.Button, QColor(20, 20, 20))
    dark_palette.setColor(QPalette.Disabled, QPalette.ButtonText, QColor(120, 120, 120))

    app.setPalette(dark_palette)

    form = HSIDialog(X)
    form.show()
    app.exec_()
    # sys.exit(app.exec_())


# If the file is run directly and not imported, this runs the main function
if __name__ == '__main__':
    hsiPlot(None)
# sys.exit(1)



import numpy as np
import hypers as hp
from hypers._tools._types import PreprocessType, PREPROCESSING_TYPES


def _data_preprocessing(X: 'hp.Dataset',
                        mdl: PreprocessType) -> None:

    if type(mdl) not in PREPROCESSING_TYPES:
        raise TypeError('Must pass a sklearn preprocessing class. Refer to documentation.')

    X.mdl_preprocess = mdl

    X_newdata = X.mdl_preprocess.fit_transform(X.flatten()).reshape(X.shape)
    X.data = X_newdata
    X.update()


def _data_scale(X: 'hp.Dataset') -> None:
    for _val in np.ndindex(X.shape[:-1]):
        X.data[_val] /= np.max(np.abs(X.data[_val]))


def _data_whiten(X: 'hp.Dataset') -> None:
    sigma = np.cov(X.flatten().T, rowvar=True)
    U, S, V = np.linalg.svd(sigma)
    epsilon = 1e-5
    zca_matrix = np.dot(U, np.dot(np.diag(1.0/np.sqrt(S + epsilon)), U.T))
    X.data = np.dot(X.flatten(), zca_matrix).reshape(X.shape)



from setuptools import setup

with open('README.md') as f:
    LONG_DESCRIPTION = f.read()

setup(
    name='hypers',

    version='0.0.10',

    packages=['hypers',
              'hypers._view',
              'hypers._view._form',
              'hypers._tools',
              'hypers._learning'],

    python_requires='>=3.5.0',

    url='https://github.com/priyankshah7/hypers',

    download_url='https://github.com/priyankshah7/hypers/archive/v0.0.10.tar.gz',

    license='BSD 3-Clause',

    author='Priyank Shah',

    author_email='priyank.shah@kcl.ac.uk',

    description='Hyperspectral data analysis and machine learning',

    long_description=LONG_DESCRIPTION,

    long_description_content_type='text/markdown',

    keywords=['hyperspectral',
              'data-analysis',
              'clustering',
              'matrix-decompositions',
              'hyperspectral-analysis',
              'machine learning'],

    install_requires=['numpy>=1.14.2',
                      'matplotlib',
                      'scipy>=1.1.0',
                      'pyqt5>=5.10.1',
                      'pyqtgraph>=0.10.0',
                      'scikit-learn>=0.20.1']
)



"""
Stores data in a custom class and generates attributes for other modules
"""
import numpy as np
from typing import Tuple, Union, List

from hypers._preprocessing import _data_preprocessing, _data_scale, _data_whiten
from hypers._learning import _data_cluster, _vca, _ucls, _data_decomposition, _data_scree, _data_mixture
from hypers._tools import _data_smoothen, _data_mean, _data_checks, _data_access
from hypers._tools import PreprocessType, ClusterType, DecomposeType, MixtureType
from hypers._view import hsiPlot


class Dataset:
    """ Dataset data structure

    Attributes
    ----------
    data: np.ndarray
        The raw hyperspectral data

    shape: tuple
        Shape of the hyperspectral data

    ndim: int
        Number of dimensions of the data (3 or 4)

    n_features: int
        Number of spectral bands

    n_samples: int
        Total number of pixels

    mean_spectrum: np.ndarray
        An array of the mean spectrum of the data

    mean_image: np.ndarray
        An array of the mean image of the data

    image: np.ndarray
        An array of the image averaged over the specified spectral bands

    spectrum: np.ndarray
        An array of the spectrum averaged over the specified pixels

    smoothing: str
        Smoothing to use. Can be either:

        - savitzky_golay
        - gaussian

        Default is 'savitzky_golay'.

    """
    def __init__(self, data: np.ndarray,
                 scale: bool = False,
                 whiten: bool = False) -> None:
        self.data = np.squeeze(data)
        self.scale = scale
        self.whiten = whiten

        # Data properties
        self.shape = None
        self.ndim = None
        self.n_features = None
        self.n_samples = None
        self.smoothing = 'savitzky_golay'

        # Hyperspectral image/spectrum
        self.image = None
        self.spectrum = None
        self.mean_image = None
        self.mean_spectrum = None

        # sklearn
        self.mdl_preprocess = None
        self.mdl_decompose = None
        self.mdl_cluster = None
        self.mdl_mixture = None

        self.update()

    def __getitem__(self, key: tuple) -> np.ndarray:
        return self.data[key]

    def __setitem__(self, key: tuple, value: Union[int, float, np.ndarray]) -> None:
        self.data[key] = value
        self.update()

    def __truediv__(self, var: Union[int, float, np.ndarray]) -> 'Dataset':
        if type(var) in (int, float):
            for _val in np.ndenumerate(self.data):
                self.data[_val[0]] /= var

        elif type(var) == np.ndarray and var.ndim == 1 and var.shape[0] == self.shape[-1]:
            for _val in np.ndindex(self.shape[:-1]):
                self.data[_val] /= var

        else:
            raise TypeError('Can only divide by an integer, float or spectral array')

        self.update()
        return self

    def __mul__(self, var: Union[int, float, np.ndarray]) -> 'Dataset':
        if type(var) in (int, float):
            for _val in np.ndenumerate(self.data):
                self.data[_val[0]] *= var

        elif type(var) == np.ndarray and var.ndim == 1 and var.shape[0] == self.shape[-1]:
            for _val in np.ndindex(self.shape[:-1]):
                self.data[_val] *= var

        else:
            raise TypeError('Can only multiply by an integer, float or spectral array')

        self.update()
        return self

    def __add__(self, var: Union[int, float, np.ndarray]) -> 'Dataset':
        if type(var) in (int, float):
            for _val in np.ndenumerate(self.data):
                self.data[_val[0]] += var

        elif type(var) == np.ndarray and var.ndim == 1 and var.shape[0] == self.shape[-1]:
            for _val in np.ndindex(self.shape[:-1]):
                self.data[_val] += var

        else:
            raise TypeError('Can only add with an integer, float or spectral array')

        self.update()
        return self

    def __sub__(self, var: Union[int, float, np.ndarray]) -> 'Dataset':
        if type(var) in (int, float):
            for _val in np.ndenumerate(self.data):
                self.data[_val[0]] -= var

        elif type(var) == np.ndarray and var.ndim == 1 and var.shape[0] == self.shape[-1]:
            for _val in np.ndindex(self.shape[:-1]):
                self.data[_val] -= var

        else:
            raise TypeError('Can only subtract by an integer, float or spectral array')

        self.update()
        return self

    def update(self) -> None:
        """Update the stored data and class properties"""
        _data_checks(self)
        if self.scale:
            _data_scale(self)
        if self.whiten:
            _data_whiten(self)
        _data_mean(self)
        _data_access(self)

    def view(self) -> None:
        """Open the hyperspectral viewer GUI"""
        hsiPlot(self)

    def smoothen(self, **kwargs) -> None:
        """Smoothen the hyperspectral data

        Parameters
        ----------
        **kwargs
            Keyword arguments for either of the following (depending on what the smoothing attribute
            has been set to):

            - ``scipy.signal.savgol_filter``
            - ``scipy.ndimage.filters.gaussian_filter``
        """
        _data_smoothen(self, **kwargs)
        self.update()

    def flatten(self) -> np.ndarray:
        """ Returns flattened array

        Returns
        -------
        x_flat: np.ndarray
            Flattened 2d array of the stored hyperspectral data
        """
        return np.reshape(self.data, (np.prod(self.shape[:-1]), self.shape[-1]))

    def scree(self, plot: bool = False,
              return_arrs: bool = True) -> np.ndarray:
        """ Returns PCA scree

        Parameters
        ----------
        plot: bool
            If True, will plot the array

        return_arrs: bool
            If True, will return the array

        Returns
        -------
        scree: np.ndarray
            An array of the PCA scree
        """
        return _data_scree(self, plot=plot, return_arrs=return_arrs)

    def preprocess(self, mdl: PreprocessType) -> None:
        """ Preprocess data

        Parameters
        ----------
        mdl: PreprocessType
            Accepts a ``scikit-learn`` preprocessing class
        """
        _data_preprocessing(self, mdl)

    def decompose(self, mdl: DecomposeType,
                  plot: bool = False,
                  return_arrs: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        """Decompose data

        Parameters
        ----------
        mdl: DecomposeType
            Accepts a ``scikit-learn`` decomposition class

        plot: bool
            If True, will return a plot of the images/spectra of the components

        return_arrs: bool
            If True, will return the arrays of the images/spectra

        Returns
        -------
        ims: np.ndarray
            An array of images (size: x, y, (z), n_components)

        specs: np.ndarray
            An array of spectra (size: spectra, n_components)
        """

        return _data_decomposition(self, mdl, return_arrs=return_arrs, plot=plot)

    def cluster(self, mdl: ClusterType,
                decomposed: bool = False,
                pca_comps: int = 4,
                plot: bool = False,
                return_arrs: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        """Cluster data

        Parameters
        ----------
        mdl: ClusterType
            Accepts a ``scikit-learn`` cluster class

        decomposed: bool
            Whether to perform PCA on the data prior to clustering

        pca_comps: int
            If decomposed=True, this specifies the number of principal components to use

        plot: bool
            If True, will return a plot of the labels/spectra of the clusters

        return_arrs: bool
            If True, will return the arrays of the labels/spectra of the clusters

        Returns
        -------
        ims: np.ndarray
            An array of the labels (size: x, y, (z))

        specs: np.ndarray
            An array of spectra (size: spectra, n_clusters)
        """

        return _data_cluster(self, mdl, decomposed=decomposed, pca_comps=pca_comps, plot=plot, return_arrs=return_arrs)

    def mixture(self, mdl: MixtureType,
                plot: bool = False,
                return_arrs: bool = True) -> Tuple[np.ndarray, np.ndarray]:
        """Gaussian mixture models

        Parameters
        ----------
        mdl: MixtureType
            Accepts a ``scikit-learn`` mixture class

        plot: bool
            If True, will return a plot of the labels/spectra of the components

        return_arrs: bool
            If True, will return the arrays of the labels/spectra of the components

        Returns
        -------
        labels: np.ndarray
            An array of the labels

        spectra: np.ndarray
            An array of the spectra of the components
        """

        return _data_mixture(self, mdl, plot=plot, return_arrs=return_arrs)

    def vca(self, n_components: int = 4,
            plot: bool = False,
            return_arrs: bool = True) -> Tuple[np.ndarray, List[int]]:
        """Vertex component analysis

        Parameters
        ----------
        n_components: int
            Number of pure components to find

        plot: bool
            If True, will return a plot of the pure spectra

        return_arrs: bool
            If True, will return an array of the pure spectra and a list of tuples of the coordinates of the
            pure pixels

        Returns
        -------
        spectra: np.ndarray
            An array of the spectra of the pure pixels

        coords: np.ndarray
            A list of tuples of the coordinates of the pure pixels
        """

        return _vca(self, n_components=n_components, plot=plot, return_arrs=return_arrs)

    def abundance(self, spectra: np.ndarray,
                  plot: bool = False,
                  return_arrs: bool = True) -> np.ndarray:
        """Abundance map with least-squares fitting

        Parameters
        ----------
        spectra: np.ndarray
            Spectra to perform fitting with

        plot: bool
            If True, will return an image of the abundance map

        return_arrs: bool
            If True, will return an array of the abundance map

        Returns
        -------
        im: np.ndarray
            An array of the abundance map
        """

        return _ucls(self, spectra=spectra, plot=plot, return_arrs=return_arrs)



import warnings
import numpy as np
import hypers as hp


def _data_checks(X: 'hp.Dataset'):
    if type(X.data) != np.ndarray:
        raise TypeError('Data must be a numpy array')

    X.shape = X.data.shape
    X.ndim = len(X.shape)

    if X.ndim != 3 and X.ndim != 4:
        raise TypeError('Data must be 3- or 4- dimensional.')

    if X.ndim == 3:
        X.n_samples = X.shape[0] * X.shape[1]
        X.n_features = X.shape[2]

        if not X.n_samples > X.n_features:
            warnings.warn('n_samples (number of pixels) should be greater than n_features (spectral points)')

    elif X.ndim == 4:
        X.n_samples = X.shape[0] * X.shape[1] * X.shape[2]
        X.n_features = X.shape[3]

        if not X.n_samples > X.n_features:
            warnings.warn('n_samples (number of pixels) should be greater than n_features (spectral points)')


def _data_mean(X: 'hp.Dataset'):
    if X.ndim == 3:
        X.mean_image = np.squeeze(np.mean(X.data, 2))
        X.mean_spectrum = np.squeeze(np.mean(np.mean(X.data, 1), 0))

    elif X.ndim == 4:
        X.mean_image = np.squeeze(np.mean(X.data, 3))
        X.mean_spectrum = np.squeeze(np.mean(np.mean(np.mean(X.data, 2), 1), 0))


def _data_access(X: 'hp.Dataset'):
    X.image = _AccessImage(X.data, X.shape, X.ndim)
    X.spectrum = _AccessSpectrum(X.data, X.shape, X.ndim)


class _AccessImage:
    def __init__(self, X, shape, ndim):
        self.data = X
        self.shape = shape
        self.ndim = ndim

    def __getitem__(self, item):
        return np.squeeze(np.mean(self.data[item], self.ndim - 1))


class _AccessSpectrum:
    def __init__(self, X, shape, ndim):
        self.data = X
        self.shape = shape
        self.ndim = ndim

    def __getitem__(self, item):
        if self.ndim == 3:
            return np.squeeze(np.mean(np.mean(self.data[item], 1), 0))

        elif self.ndim == 4:
            return np.squeeze(np.mean(np.mean(np.mean(self.data[item], 2), 1), 0))
